@@ -14,13 +14,11 @@
  * and Design Engineering (MADE) Initiative through ARPA contract
  * F33615-94-C-4400.
  *
- * SCCS: @(#) tclLoadAout.c 1.9 97/02/22 14:05:01
+ * RCS: @(#) $Id: tclLoadAout.c,v 1.3 1999/04/16 00:48:04 stanton Exp $
  */
 
 #include "tclInt.h"
-#include "tclPort.h"
 #include <fcntl.h>
-#include "../compat/dlfcn.h"
 #ifdef HAVE_EXEC_AOUT_H
 #   include <sys/exec_aout.h>
 #endif
@@ -82,66 +80,34 @@ extern char * sbrk _ANSI_ARGS_((size_t));
 
 static char * SymbolTableFile = NULL;
 
+/*
+ * Type of the dictionary function that begins each load module.
+ */
+
+typedef Tcl_PackageInitProc * (* DictFn) _ANSI_ARGS_ ((char * symbol));
+
 /*
  * Prototypes for procedures referenced only in this file:
  */
 
-static VOID * LoadDictionary _ANSI_ARGS_((char *symbol));
-static int FindLibraries _ANSI_ARGS_((CONST char *fileName, Tcl_DString *buf));
+static int FindLibraries _ANSI_ARGS_((Tcl_Interp * interp, char * fileName,
+				      Tcl_DString * buf));
 static void UnlinkSymbolTable _ANSI_ARGS_((void));
-static void Seterror _ANSI_ARGS_((char *message));
-
-/*
- *----------------------------------------------------------------------
- *
- * LoadDictionary --
- *
- * Search for functions in Tcl library.
- *
- * Results:
- *	addres of symbol, or NULL if it cannot be found.
- *	Currently, only the Tcl_Init() function can be found.
- *
- * Side effects:
- *	None.
- *
- *----------------------------------------------------------------------
- */
-
-static struct {
-  char * name;
-  int (*value) _ANSI_ARGS_((Tcl_Interp *));
-} dictionary[] = {
-  { "Tcl_Init", Tcl_Init },
-  0, 0 
-};
-
-static VOID *
-LoadDictionary (symbol)
-    char * symbol;
-{
-    int i;
-    for (i = 0; dictionary[i] . name != 0; ++i) {
-      if (!strcmp (symbol, dictionary[i] . name)) {
-	return (VOID *) dictionary[i].value;
-      }
-    }
-    return (VOID *) NULL;
-}
-
-static char *errorMessage = NULL;
-
 
 /*
  *----------------------------------------------------------------------
  *
- * dlopen --
+ * TclpLoadFile --
  *
- *	Dynamically loads a binary code file into memory.
+ *	Dynamically loads a binary code file into memory and returns
+ *	the addresses of two procedures within that file, if they
+ *	are defined.
  *
  * Results:
- *	A handle which can be used in later calls to dlsym(),
- *	or NULL when the attempt fails.
+ *	A standard Tcl completion code.  If an error occurs, an error
+ *	message is left in the interp's result.  *proc1Ptr and *proc2Ptr
+ *	are filled in with the addresses of the symbols given by
+ *	*sym1 and *sym2, or NULL if those symbols can't be found.
  *
  * Side effects:
  *	New code suddenly appears in memory.
@@ -169,10 +135,19 @@ static char *errorMessage = NULL;
  *----------------------------------------------------------------------
  */
 
-VOID *
-dlopen(path, flags)
-    CONST char *path;
-    int flags;
+int
+TclpLoadFile(interp, fileName, sym1, sym2, proc1Ptr, proc2Ptr, clientDataPtr)
+    Tcl_Interp *interp;		/* Used for error reporting. */
+    char *fileName;		/* Name of the file containing the desired
+				 * code (UTF-8). */
+    char *sym1, *sym2;		/* Names of two procedures to look up in
+				 * the file's symbol table. */
+    Tcl_PackageInitProc **proc1Ptr, **proc2Ptr;
+				/* Where to return the addresses corresponding
+				 * to sym1 and sym2. */
+    ClientData *clientDataPtr;	/* Filled with token for dynamically loaded
+				 * file which will be passed back to 
+				 * TclpUnloadFile() to unload the file. */
 {
   char * inputSymbolTable;	/* Name of the file containing the 
 				 * symbol table from the last link. */
@@ -182,68 +157,30 @@ dlopen(path, flags)
   char relocatedFileName [L_tmpnam];
 				/* Name of the file holding the relocated */
 				/* text of the module */
-  int relocatedFd = -1;		/* File descriptor of the file holding
+  int relocatedFd;		/* File descriptor of the file holding
 				 * relocated text */
   struct exec relocatedHead;	/* Header of the relocated text */
   unsigned long relocatedSize;	/* Size of the relocated text */
   char * startAddress;		/* Starting address of the module */
+  DictFn dictionary;		/* Dictionary function in the load module */
   int status;			/* Status return from Tcl_ calls */
-  char *p;
-  CONST char *r, *pkgGuess;
-  Tcl_Interp *interp = NULL;
-  Tcl_DString fullPath;
-
-  if (path == (char *) NULL) {
-    return (VOID *) LoadDictionary;
-  }
-
-  errno = 0;
-  if (errorMessage) {
-    ckfree(errorMessage);
-    errorMessage = NULL;
-  }
+  char * p;
 
+  *clientDataPtr = NULL;
+  
   /* Find the file that contains the symbols for the run-time link. */
 
   if (SymbolTableFile != NULL) {
     inputSymbolTable = SymbolTableFile;
   } else if (tclExecutableName == NULL) {
-    Seterror("can't find the tclsh executable");
-    goto error;
+    Tcl_SetResult (interp, "can't find the tclsh executable", TCL_STATIC);
+    return TCL_ERROR;
   } else {
     inputSymbolTable = tclExecutableName;
   }
 
   /* Construct the `ld' command that builds the relocated module */
 
-  interp = Tcl_CreateInterp();
-
-  Tcl_DStringInit (&fullPath);
-  if (Tcl_GetPathType(path) == TCL_PATH_RELATIVE) {
-	p = getenv("LD_LIBRARY_PATH");
-	while (p) {
-	    if ((q = strchr(p,':')) == NULL) {
-		q = p; while(*q) q++;
-	    }
-	    if (p == q) break;
-	    Tcl_DStringAppend(&fullPath, p, q-p);
-	    Tcl_DStringAppend(&fullPath, "/", 1);
-	    Tcl_DStringAppend(&fullPath, path, -1);
-	    if (access(Tcl_DStringValue(&fullPath), F_OK) != -1) {
-		break;
-	    }
-	    Tcl_DStringSetLength(&fullPath, 0);
-	    p = q; if (*p) p++;
-	}
-	if (!q && (access(path, F_OK) == -1)) {
-	    Tcl_DStringAppend(&fullPath, LIB_RUNTIME_DIR, -1);
-	    Tcl_DStringAppend(&fullPath, "/", 1);
-	    Tcl_DStringAppend(&fullPath, path, -1);
-	}
-  }
-  if (*Tcl_DStringValue(&fullPath) == 0) {
-    Tcl_DStringAppend(&fullPath, path, -1);
-  }
   tmpnam (relocatedFileName);
   Tcl_DStringInit (&linkCommandBuf);
   Tcl_DStringAppend (&linkCommandBuf, "exec ld -o ", -1);
@@ -252,58 +189,16 @@ dlopen(path, flags)
   Tcl_DStringAppend (&linkCommandBuf, " -G 0 ", -1);
 #endif
   Tcl_DStringAppend (&linkCommandBuf, " -u TclLoadDictionary_", -1);
-  if (pkgGuess = strrchr(path,'/')) {
-    pkgGuess++;
-  } else {
-    pkgGuess = path;
-  }
-  if (!strncmp(pkgGuess,"lib",3)) {
-    pkgGuess+=3;
-  }
-  for (r = pkgGuess; (*r) && (*r != '.'); r++) {
-    /* Empty loop body. */
-  }
-  if ((r>pkgGuess+3) && !strncmp(r-3,"_G0.",4)) {
-    r-=3;
-  }
-  while ((r-- > pkgGuess) && isdigit(UCHAR(*r))) {
-    /* Empty loop body. */
-  }
-  r++;
-  Tcl_DStringAppend(&linkCommandBuf,(char *) pkgGuess, r-pkgGuess);
-
-  p = Tcl_DStringValue(&linkCommandBuf);
-  p += strlen(p) - (r-pkgGuess);
-
-  if (islower(UCHAR(*p))) {
-    *p = (char) toupper(UCHAR(*p));
-  }
-  while (*(p++)) {
-    if (isupper(UCHAR(*p))) {
-	*p = (char) tolower(UCHAR(*p));
-    }
-  }
+  TclGuessPackageName(fileName, &linkCommandBuf);
   Tcl_DStringAppend (&linkCommandBuf, " -A ", -1);
   Tcl_DStringAppend (&linkCommandBuf, inputSymbolTable, -1);
   Tcl_DStringAppend (&linkCommandBuf, " -N -T XXXXXXXX ", -1);
-  Tcl_DStringAppend (&linkCommandBuf, Tcl_DStringValue(&fullPath), -1);
-  p = getenv("LD_LIBRARY_PATH");
-  while (p) {
-    if ((q = strchr(p,':')) == NULL) {
-      q = p; while(*q) q++;
-    }
-    if (p == q) break;
-    Tcl_DStringAppend(&linkCommandBuf, " -L", 3);
-    Tcl_DStringAppend(&linkCommandBuf, p, q-p);
-    p = q; if (*p) p++;
-  }
+  Tcl_DStringAppend (&linkCommandBuf, fileName, -1);
   Tcl_DStringAppend (&linkCommandBuf, " ", -1);
-  if (FindLibraries (Tcl_DStringValue(&fullPath), &linkCommandBuf) != TCL_OK) {
+  if (FindLibraries (interp, fileName, &linkCommandBuf) != TCL_OK) {
     Tcl_DStringFree (&linkCommandBuf);
-    Tcl_DStringFree (&fullPath);
-    goto error;
+    return TCL_ERROR;
   }
-  Tcl_DStringFree (&fullPath);
   linkCommand = Tcl_DStringValue (&linkCommandBuf);
 
   /* Determine the starting address, and plug it into the command */
@@ -320,9 +215,7 @@ dlopen(path, flags)
   status = Tcl_Eval (interp, linkCommand);
   Tcl_DStringFree (&linkCommandBuf);
   if (status != 0) {
-    Seterror(interp->result);
-    errno = 0;
-    goto error;
+    return TCL_ERROR;
   }
 
   /* Open the linker's result file and read the header */
@@ -339,14 +232,16 @@ dlopen(path, flags)
   /* Check the magic number */
 
   if (relocatedHead.a_magic != OMAGIC) {
-    Seterror("bad magic number in intermediate file");
+    Tcl_AppendResult (interp, "bad magic number in intermediate file \"",
+		      relocatedFileName, "\"", (char *) NULL);
     goto failure;
   }
 
   /* Make sure that memory allocation is still consistent */
 
   if ((unsigned long) sbrk (0) > (unsigned long) startAddress) {
-    Seterror("can't load, memory allocation is inconsistent");
+    Tcl_SetResult (interp, "can't load, memory allocation is inconsistent.",
+		   TCL_STATIC);
     goto failure;
   }
 
@@ -355,7 +250,7 @@ dlopen(path, flags)
   relocatedSize = relocatedHead.a_text + relocatedHead.a_data
     + relocatedHead.a_bss;
   if (relocatedSize > TCL_LOADMAX) {
-    Seterror("module too big to load");
+    Tcl_SetResult (interp, "module too big to load", TCL_STATIC);
     goto failure;
   }
 
@@ -382,10 +277,12 @@ dlopen(path, flags)
   if (read (relocatedFd, startAddress, relocatedSize) < relocatedSize) {
     brk (startAddress);
   ioError:
-    Seterror("error on intermediate file: ");
+    Tcl_AppendResult (interp, "error on intermediate file \"",
+		      relocatedFileName, "\": ", Tcl_PosixError (interp),
+		      (char *) NULL);
   failure:
     (void) unlink (relocatedFileName);
-    goto error;
+    return TCL_ERROR;
   }
 
   /* Close the intermediate file. */
@@ -393,132 +290,24 @@ dlopen(path, flags)
   (void) close (relocatedFd);
 
   /* Arrange things so that intermediate symbol tables eventually get
-   * deleted. If the flag RTLD_GLOBAL is not set, just keep the
-   * old file. */
+   * deleted. */
 
-  if (flags & RTLD_GLOBAL) {
-    if (SymbolTableFile != NULL) {
-      UnlinkSymbolTable ();
-    } else {
-      atexit (UnlinkSymbolTable);
-    }
-    SymbolTableFile = ckalloc (strlen (relocatedFileName) + 1);
-    strcpy (SymbolTableFile, relocatedFileName);
+  if (SymbolTableFile != NULL) {
+    UnlinkSymbolTable ();
   } else {
-    (void) unlink (relocatedFileName);
+    atexit (UnlinkSymbolTable);
   }
-  return (VOID *) startAddress;
+  SymbolTableFile = ckalloc (strlen (relocatedFileName) + 1);
+  strcpy (SymbolTableFile, relocatedFileName);
+  
+  /* Look up the entry points in the load module's dictionary. */
 
-error:
-  if (relocatedFd>=0) {
-	close (relocatedFd);
-  }
-  if (interp) {
-    Tcl_DeleteInterp(interp);
-  }
-  return NULL;
+  dictionary = (DictFn) startAddress;
+  *proc1Ptr = dictionary (sym1);
+  *proc2Ptr = dictionary (sym2);
+
+  return TCL_OK;
 }
-
-
-/*
- *----------------------------------------------------------------------
- *
- * dlsym --
- *
- *	This function returns the address of a
- *	symbol, give the handle returned by dlopen().
- *
- * Results:
- *	Returns the address of the symbol in the dll.
- *
- * Side effects:
- *	None.
- *
- *----------------------------------------------------------------------
- */
-
-VOID *dlsym(handle, symbol)
-    VOID *handle;
-    CONST char *symbol;
-{
-    if ((handle != NULL) && (symbol != NULL)) {
-	return ((VOID * (*) _ANSI_ARGS_((CONST char *))) handle) (symbol);
-    } else {
-	return (VOID *) NULL;
-    }
-}
-
-
-/*
- *----------------------------------------------------------------------
- *
- * dlerror --
- *
- *	This function returns a string describing the error which
- *	occurred in dlopen().
- *
- * Results:
- *	Returns an error message.
- *
- * Side effects:
- *	None.
- *
- *----------------------------------------------------------------------
- */
-
-char *
-dlerror()
-{
-    char *err, *msg;
-
-    if (errorMessage && errno) {
-	err = Tcl_ErrnoMsg(errno);
-	msg = ckalloc(strlen(errorMessage)+strlen(err)+1);
-	strcpy(msg, errorMessage);
-	strcat(msg, err);
-	ckfree(errorMessage);
-	errorMessage = msg;
-    }
-    return errorMessage;
-}
-
-
-/*
- *----------------------------------------------------------------------
- *
- * dlclose --
- *
- *	Just a dummy function, only for compatibility. There is no
- *	way to remove dll's from memory.
- *
- * Results:
- *	Always returns 0 (= O.K.)
- *
- * Side effects:
- *	None
- *
- *----------------------------------------------------------------------
- */
-
-int
-dlclose(handle)
-    VOID *handle;
-{
-    return 0;
-}
-
-static void
-Seterror(message)
-    char *message;
-{
-    if (errorMessage) {
-	ckfree(errorMessage);
-    }
-    errorMessage = ckalloc(strlen(message)+1);
-    strcpy(errorMessage, message);
-    return;
-}
-
 
 /*
  *------------------------------------------------------------------------
@@ -529,25 +318,33 @@ Seterror(message)
  *
  * Results:
  *	A standard Tcl completion code.  If an error occurs,
- *	an error message is left in interp->result.  The -l and -L flags
- *	are concatenated onto the dynamic string `buf'.
+ *	an error message is left in the interp's result.  The -l and -L
+ *	flags are concatenated onto the dynamic string `buf'.
  *
  *------------------------------------------------------------------------
  */
 
 static int
-FindLibraries (fileName, buf)
-     CONST char * fileName;	/* Name of the load module */
+FindLibraries (interp, fileName, buf)
+     Tcl_Interp * interp;	/* Used for error reporting */
+     char * fileName;		/* Name of the load module */
      Tcl_DString * buf;		/* Buffer where the -l an -L flags */
 {
   FILE * f;			/* The load module */
-  int c = EOF;			/* Byte from the load module */
+  int c;			/* Byte from the load module */
   char * p;
+  Tcl_DString ds;
+  CONST char *native;
 
   /* Open the load module */
 
-  if ((f = fopen (fileName, "rb")) == NULL) {
-    Seterror("");
+  native = Tcl_UtfToExternalDString(NULL, fileName, -1, &ds);
+  f = fopen(native, "rb");				/* INTL: Native. */
+  Tcl_DStringFree(&ds);
+  
+  if (f == NULL) {
+    Tcl_AppendResult (interp, "couldn't open \"", fileName, "\": ",
+		      Tcl_PosixError (interp), (char *) NULL);
     return TCL_ERROR;
   }
 
@@ -566,11 +363,13 @@ FindLibraries (fileName, buf)
     }
   }
 
-  /* No library list -- assume no dependancies */
+  /* No library list -- this must be an ill-formed module */
 
   if (c == EOF) {
+    Tcl_AppendResult (interp, "File \"", fileName,
+		      "\" is not a Tcl load module.", (char *) NULL);
     (void) fclose (f);
-    return TCL_OK;
+    return TCL_ERROR;
   }
 
   /* Accumulate the library list */
@@ -582,7 +381,8 @@ FindLibraries (fileName, buf)
   (void) fclose (f);
 
   if (c == EOF) {
-    Seterror("Library directory ends prematurely");
+    Tcl_AppendResult (interp, "Library directory in \"", fileName,
+		      "\" ends prematurely.", (char *) NULL);
     return TCL_ERROR;
   }
 
@@ -614,3 +414,94 @@ UnlinkSymbolTable ()
   ckfree (SymbolTableFile);
   SymbolTableFile = NULL;
 }
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * TclpUnloadFile --
+ *
+ *	Unloads a dynamically loaded binary code file from memory.
+ *	Code pointers in the formerly loaded file are no longer valid
+ *	after calling this function.
+ *
+ * Results:
+ *	None.
+ *
+ * Side effects:
+ *	Does nothing.  Can anything be done?
+ *
+ *----------------------------------------------------------------------
+ */
+
+void
+TclpUnloadFile(clientData)
+    ClientData clientData;	/* ClientData returned by a previous call
+				 * to TclpLoadFile().  The clientData is 
+				 * a token that represents the loaded 
+				 * file. */
+{
+}
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * TclGuessPackageName --
+ *
+ *	If the "load" command is invoked without providing a package
+ *	name, this procedure is invoked to try to figure it out.
+ *
+ * Results:
+ *	Always returns 0 to indicate that we couldn't figure out a
+ *	package name;  generic code will then try to guess the package
+ *	from the file name.  A return value of 1 would have meant that
+ *	we figured out the package name and put it in bufPtr.
+ *
+ * Side effects:
+ *	None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+int
+TclGuessPackageName(fileName, bufPtr)
+    char *fileName;		/* Name of file containing package (already
+				 * translated to local form if needed). */
+    Tcl_DString *bufPtr;	/* Initialized empty dstring.  Append
+				 * package name to this if possible. */
+{
+    char *p, *q, *r;
+    int srcOff, dstOff;
+
+    if (q = strrchr(fileName,'/')) {
+	q++;
+    } else {
+	q = fileName;
+    }
+    if (!strncmp(q,"lib",3)) {
+	q+=3;
+    }
+    p = q;
+    while ((*p) && (*p != '.') && ((*p<'0') || (*p>'9'))) {
+	p++;
+    }
+    if ((p>q+2) && !strncmp(p-2,"_G0.",4)) {
+	p-=2;
+    }
+    if (p<q) {
+	return 0;
+    }
+
+    Tcl_DStringAppend(bufPtr,q, p-q);
+
+    r = Tcl_DStringValue(bufPtr);
+    r += strlen(r) - (p-q);
+
+    /*
+     * Capitalize the string and then recompute the length.
+     */
+
+    Tcl_UtfToTitle(r);
+    Tcl_DStringSetLength(bufPtr, strlen(Tcl_DStringValue(bufPtr)));
+
+    return 1;
+}
